﻿namespace Mandelbrot

[<AutoOpen>]
module Computation =
    open System
    open Microsoft.FSharp.Math

    let maxIteration = 255
    let modSquared (c : Complex) = 
        c.RealPart * c.RealPart + c.ImaginaryPart * c.ImaginaryPart
    
    let (|Escaped|DidNotEscape|) c =
        let rec compute z iterations =
            if(modSquared z >= 4.0) 
                then Escaped iterations
            elif iterations = maxIteration
                then DidNotEscape
            else compute ((z * z) + c) (iterations + 1)
        compute c 0

open System.Windows
open System.Windows.Controls
open System.Windows.Media
open System.Windows.Media.Imaging
open Utils

type ZoomControl(width:int,height:int) as control =
    inherit UserControl()
    let bitmap = WriteableBitmap(width,height)
    let canvas = Canvas()
    do  canvas.Children.Add(Image(Source=bitmap))
    do  control.Content <- canvas 
    let render (x1,y1,x2,y2) (offset, height, stripe) (pixels:int[]) =
        let dx, dy = x2-x1, y2-y1
        for y = 0 to height-1 do
            for x = 0 to width-1 do
                let y = offset + (y*stripe)
                pixels.[x+y*width] <-
                    let x = ((float x/float width) * dx) + x1
                    let y = ((float y/float (height*stripe)) * dy) + y1 
                    match Complex.Create(x, y) with
                    | DidNotEscape -> 0xff000000
                    | Escaped i -> 
                        let i = 255 - i
                        0xff000000 + (i <<< 16) +  (i <<< 8) + i
    
    let mutable points = (-2.0, -1.0, 1.0, 1.0)
    do  render points (0,height,1) bitmap.Pixels

    let copy (l,t,w,h) =
        let selection = WriteableBitmap(int w, int h)
        let source = bitmap.Pixels
        let dest = selection.Pixels
        for y = 0 to int h - 1 do
            for x = 0 to int w - 1 do
                dest.[x + (y * int w)] <- 
                    source.[int l + x + ((int t + y) * width)]
        selection

    let moveControl (element:FrameworkElement) (start:Point) (finish:Point) =
        element.Width <- abs(finish.X - start.X)
        element.Height <- abs(finish.Y - start.Y)
        Canvas.SetLeft(element, min start.X finish.X)
        Canvas.SetTop(element, min start.Y finish.Y)

    let transparentGray = 
        SolidColorBrush(Color.FromArgb(128uy, 164uy, 164uy, 164uy))

    let buffer = Array.create (width * height) 0

    let rec waiting() = async {
        let! md = Async.AwaitObservable(control.MouseLeftButtonDown)
        let rc = new Canvas(Background = transparentGray)
        canvas.Children.Add(rc) 
        do! drawing(rc, md.GetPosition(canvas)) }

    and drawing(rc:Canvas, pos) = async {
        let! evt = Async.AwaitObservable(canvas.MouseLeftButtonUp, canvas.MouseMove)
        match evt with
        | Choice1Of2(up) ->
            let l, t = Canvas.GetLeft(rc), Canvas.GetTop(rc)
            let w, h = rc.Width, rc.Height
            if w > 1.0 && h > 1.0 then
                let preview = 
                    Image(Source=copy (l,t,w,h),
                          Stretch=Stretch.Fill,
                          Width=float width,Height=float height)
                canvas.Children.Add preview

                let zoom (x1,y1,x2,y2) =           
                    let tx x = ((x/float width) * (x2-x1)) + x1
                    let ty y = ((y/float height) * (y2-y1)) + y1
                    tx l, ty t, tx (l+w), ty (t+h)

                points <- zoom points
                 
                do! [0..3] 
                    |> List.map (fun y -> async {
                        render points (y,(height/4),4) buffer
                    })
                    |> Async.Parallel
                    |> Async.Ignore

                Array.blit buffer 0 bitmap.Pixels 0 (width*height)

                canvas.Children.Remove preview |> ignore

            canvas.Children.Remove rc |> ignore
            do! waiting() 
        | Choice2Of2(move) ->
            moveControl rc pos (move.GetPosition(canvas))
            do! drawing(rc, pos) }
    
    do  waiting() |> Async.StartImmediate

type App() as app =
    inherit System.Windows.Application()
    do  app.Startup.Add(fun _ -> app.RootVisual <- ZoomControl(512,384))